library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

use work.CamControl;
use work.EdgeDet;
use work.MemPack;
use work.CamInit;

entity Kamera is
	Port ( 
		-- fpga Takt
		clk : in std_logic;
			
		-- serial port
		rs232RX : in  std_logic;
		rs232TX : out  std_logic;
	
		-- camera pins
		XCLK : out std_logic; -- system clock
		RSTB : out std_logic := '0'; -- reset (active low)
		PWDN : out std_logic; -- power down (power down = high)
			
		SCL : out std_logic; -- I2C clock
		SDA1 : inout std_ulogic; -- I2C i/o
		SDA2 : inout std_ulogic; -- I2C i/o
		
		PCLK1 : in std_logic; -- pixel clock cam1
		PCLK2 : in std_logic; -- pixel clock cam2
		HREF1 : in std_logic;
		HREF2 : in std_logic;
		camVSYNC1 : in std_logic;
		camVSYNC2 : in std_logic;
		D : in std_logic_vector (1 downto 0); -- unbenutzt
		D1 : in std_logic_vector(9 downto 0); -- pic data cam1
		D2 : in std_logic_vector(9 downto 0); -- pic data cam2
		
		-- monitor pins
		r : out std_logic_vector(7 downto 0);
		g : out std_logic_vector(7 downto 0);
		b : out std_logic_vector(7 downto 0);
		m1 : out std_logic;
		m2 : out std_logic;
		mclk : out std_logic;
		syncT : out std_logic;
		nsync : out std_logic;
		nblank : out std_logic;
		mVSYNC : out std_logic;
		mHSYNC : out std_logic;
		
		-- sram
		sr1ce : out std_logic; -- chip enable
		sr1ub : out std_logic; -- upper byte
		sr1lb : out std_logic; -- lower byte
		sr2ce : out std_logic;
		sr2ub : out std_logic;
		sr2lb : out std_logic;
		sr12we : out std_logic := '1';
		sr12oe : out std_logic;
		sraddr : out std_logic_vector(17 downto 0);
		srio : inout std_logic_vector(31 downto 0);
		
		-- debug
		led : out std_logic_vector(7 downto 0);
		sw : in std_logic_vector(7 downto 0);
		btn : in std_logic_vector(3 downto 0);
		
		sigled : out std_logic_vector(3 downto 0);
		
		seg7 : out std_logic_vector(7 downto 0);
		seg7de : out std_logic_vector(3 downto 0)

	);
end Kamera;


architecture Behavioral of Kamera is

	-- VGA output registers
	signal curPixelPosX: std_logic_vector(15 downto 0) := x"0000";
	signal curPixelPosY: std_logic_vector(15 downto 0) := x"0000";
	
	signal h_Sync: std_logic;
	signal v_Sync: std_logic;
	
	signal hCounter: integer range 0 to 799:=0;
	signal vCounter: integer range 0 to 520:=0;
	signal hBlack: std_logic:='1';
	signal vBlack: std_logic:='1';

	type tVGAState is (VGA_Output, Capture1, Capture2, Capture3);
	signal VGAState : tVGAState := VGA_Output;

	-- clock
	signal dcmLocked : std_logic;
	signal clk1x : std_logic;
	signal clk2x : std_logic;

	signal clk25: std_logic;
	signal clkxclk: std_logic_vector(31 downto 0);
	signal clk2: std_logic_vector(31 downto 0);

	-- SRAM Adress-Zeiger
	constant SramInpDir : std_logic_vector(31 downto 0) := "ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ";

	signal rPos : std_logic_vector(17 downto 0) := "000000000000000000";
	signal rPosTmp : std_logic_vector(17 downto 0) := "000000000000000000";
	
	signal wPos : std_logic_vector(17 downto 0) := "000000000000000000";
	
	-- Kamera Signalabtastung
	signal dinBuf : std_logic_vector(7 downto 0);
	signal dinBufPrev : std_logic_vector(7 downto 0);
	signal hrefBuf : EdgeDet.tBuffer := EdgeDet.tBuffer_InitLow;
	signal vsyncBuf : EdgeDet.tBuffer := EdgeDet.tBuffer_InitHigh;
	signal pclkBuf : EdgeDet.tBuffer := EdgeDet.tBuffer_InitHigh;
	signal SDABuf : std_logic;
	
	signal isFrame : boolean := False;
	
	-- debug
	signal btnBuf : EdgeDet.tBuffer := EdgeDet.tBuffer_InitLow;
	signal ledBuf : std_logic_vector(7 downto 0);
	signal sigledBuf : std_logic_vector(3 downto 0);
	
	signal captureReset : std_logic;
	
	-- SRAM <-> Kamera Abtast Synchronisation
	signal packPos : integer range 0 to 3 := 0;
	signal writeRequest : boolean := false;
	signal writeDiffRequest : boolean := false;
	signal writeData : std_logic_vector(31 downto 0);
	
	signal readRequest : boolean := false;
	signal readLastRequest : boolean := false;
	
	-- blinking led phase config
	signal pictCnt : std_logic_vector(31 downto 0);
	signal pictRef : std_logic_vector(31 downto 0) := x"00000000";
	signal btnSampleClk : std_logic;
	signal seg7data : std_logic_vector(15 downto 0);
	
	signal cam_init : std_logic;
	signal rsrcv_int : std_logic;
	signal rs232TX_int : std_logic;
	signal rs232_used : std_logic;
	signal rs232RX_int : std_logic;

begin

	-- feste Ausgaben

	led <= ledBuf;
	
	sr1ce <= '0'; sr2ce <= '0';
	sr1ub <= '0'; sr2ub <= '0';
	sr1lb <= '0'; sr2lb <= '0';
	sr12oe <= '0';
	
	sigled <= sigledBuf;

	-- Takterzeugung 100 MHz

	clk2x_dcm_inst : entity clk2x_dcm(BEHAVIORAL) port map (
		CLKIN_IN => clk, -- unbenutzt
		RST_IN => '0',
		CLKDV_OUT => open, -- unbenutzt
		CLKIN_IBUFG_OUT => open,
		CLK0_OUT => clk1x, -- 50 MHz
		CLK2X_OUT => clk2x, -- 100 MHz
		LOCKED_OUT => dcmLocked
	);
	
	-- seg7 display
	seg7_inst : entity Seg7Display(Behavioral) port map (
		Clk => clk1x,
		Seg7 => seg7,
		Seg7DE => seg7de,
		Data => seg7data
	);
	
	process(clk1x)
		variable Q : std_logic_vector(31 downto 0) := x"00000000";
	begin
		if clk1x'event and clk1x='1' then
			btnSampleClk <= Q(16);
			Q := Q + 1;
			clk25 <= not clk25; -- VGA-Takt 
			mclk <= clk25;
			clkxclk <= clkxclk + '1'; -- eigener Takt fuer XCLK (s. unten, ca 6 MHz)
		end if;
	end process;
	
	-- Abtastprocess fuer Buttons
	
	process(btnSampleClk)
	begin
		if btnSampleClk'event and btnSampleClk='1' then
			if btn(0)='1' then
				pictRef <= pictRef + x"100";
			elsif btn(1)='1' then
				pictRef <= pictRef - x"100";
			elsif btn(3)='1' then
				cam_init <= '1';
			elsif btn(3)='0' then
				cam_init <= '0';
			end if;
		end if;
	end process;

	-- ###########################################
	-- ### monitor output
	-- ###########################################

	-- VGA DAC control
	m1 <= '0';
	m2 <= '0';
	nblank <= '1';
	nsync <= '1';
	syncT <= '0';
	mVSYNC <= v_Sync;
	mHSYNC <= h_Sync;

	process(clk2x)
			variable newline : boolean;
			
			variable tmpByte : std_logic_vector(7 downto 0);
			
			variable isLine : boolean := False;
			variable writeDataTmp : std_logic_vector(31 downto 0);
			variable readDataTmp : std_logic_vector(31 downto 0);

	begin
		if clk2x'event and clk2x='1' and dcmLocked='1' then

			-- ### VGA output ### (alle vier Zustaende)
			
			case VGAState is
			
				when VGA_Output => -- jeder 4. Zustand bei 100 MHz => 25 MHz VGA Takt

					-- horizontal control
					hCounter <= hCounter + 1; -- Spaltenzaehler + 1
					newline:=False; -- keine neue Zeile
					case hCounter is
						when  95 => h_Sync <= '0';  -- Zeilen-Sync aus
										curPixelPosX <= x"0000";
						when 143 => hBlack <= '0'; -- schwarzer Rand aus
						-- zwischen 143 und 782 liegen die sichtbaren 640 Spalten
						when 783 => hBlack <= '1'; -- schwarzer Rand ein
						when 799 => hCounter <= 0; -- newline
								 newline := True; -- neue Zeile
								 h_Sync <= '1';  -- Zeilen-Sync ein
								 rPos <= curPixelPosY * conv_std_logic_vector(160, 18);
								 
								 -- 640 / 4 = 160 : Werte je 8 Bit, Leseposition springt in Y-Richtung immer
								 -- um 160 Adressen weiter. Vier Datenbloecke werden immer zusammen unter
								 -- einer Adresse gespeichert.
								 
						when others => null;
					end case;
				
					-- vertical control
					if newline then -- nur bei neuer Zeile
						vCounter <= vCounter + 1; -- Zeilenzaehler + 1
						curPixelPosY <= curPixelPosY + x"0001"; 
						case	vCounter is
							when   0 => v_Sync <= '1';
							when   2 => v_Sync <= '0';
							when  33 => vBlack <= '0';
											curPixelPosY <= x"0000"; 
											rPos <= "000000000000000000";
							-- zwischen 33 und 509 liegen 476 sichtbare Zeilen
							when 514 => vBlack<='1';
							when 520 => vCounter<=0;
											ledBuf(5) <= not ledBuf(5);
							when others => null;
						end case;
					end if;

					-- Neue Daten einlesen, ansonsten letztes Bild anzeigen
					if readRequest then
						readDataTmp := srio;
						readRequest <= false;
					elsif readLastRequest then
--						writeDataLastFrame <= srio;
						readLastRequest <= false;
					end if;

					-- Default "schwarzes Pixel" wenn unten nicht ueberschrieben
					
					r <= x"00";
					g <= x"00";
					b <= x"00";

					if hBlack='0' and vBlack='0' then -- sichtbarer Bildbereich

						curPixelPosX <= curPixelPosX + x"0001";
						
						-- Bayer-Pattern:
						
						-- wenn:						|	00		01		10		11
						--								------------------------
						-- wenn PosY=0, dann:	|	G		R		G		R
						-- sonst:					|	B		G		B		G

						case curPixelPosX(1 downto 0) is
							when "00" =>
								tmpByte := readDataTmp(7 downto 0);
								if curPixelPosY(0) = '0' then -- gerade Zeile
									g <= tmpByte;
								else -- ungerade Zeile
									b <= tmpByte;
								end if;
							when "01" =>
								tmpByte := readDataTmp(15 downto 8);
								if curPixelPosY(0) = '0' then
									r <= tmpByte;
								else
									g <= '0' & tmpByte(7 downto 1);
								end if;
							when "10" =>
								tmpByte := readDataTmp(23 downto 16);
								if curPixelPosY(0) = '0' then
									g <= '0' & tmpByte(7 downto 1);
								else
									b <= tmpByte;
								end if;
							when "11" =>
								tmpByte := readDataTmp(31 downto 24);
								if curPixelPosY(0) = '0' then
									r <= tmpByte;
								else
									g <= '0' & tmpByte(7 downto 1);
								end if;
								rPos <= rPos + 1;
								readRequest <= true;	-- neue Daten einlesen
							when others => null;
						end case;
					
					end if;
					
					-- Adresse zum Schreiben setzen
					if writeRequest then
						sraddr <= wPos;
					end if;
					
					VGAState <= Capture1;
				
				when Capture1 =>
					

					-- in den SRAM schreiben
					if writeRequest then
						srio <= writeData; -- Daten an SRAM
						sr12we <= '0'; -- Schreiben aktivieren
						wPos <= wPos + 1; -- Position fuer den naechsten Schreibzugriff erhoehenincrement write pos for next write
						writeRequest <= false;
						readLastRequest <= true;
					end if;
	
					rPosTmp <= rPos;
					
					VGAState <= Capture2;
				
				when Capture2 =>
					sr12we <= '1'; -- Schreiben deaktivieren

					if readRequest then
						-- Ausgabeadresse zum Lesen fr VGA Ausgabe:
						-- 10 bit fuer Spalte, 8 bit fuer Zeile
						sraddr <= rPosTmp;
						-- i/o input direction
						srio <= SramInpDir;
					elsif readLastRequest then
						sraddr <= wPos;
						srio <= SramInpDir;
					end if;
					
					VGAState <= Capture3;
				
				when Capture3 =>
				
					VGAState <= VGA_Output; -- zum naechsten Zustand wechseln
					
			end case;
			
			-- ************************************************************************************+
			
			pictCnt <= pictCnt + 1; -- Bildzaehler erhoehen
			
			-- Schreibposition zu Beginn immer auf 0
			if EdgeDet.risingEdge(vsyncBuf) then
				isFrame <= False; -- nicht sichtbarer Bereich
				wPos <= "000000000000000000";

			-- Bild beginnt
			elsif EdgeDet.fallingEdge(vsyncBuf) then
				isFrame <= True; -- sichtbarer Bereich vertikal setzen
				pictCnt <= x"00000000";
			end if;
			
			if isFrame then -- im sichtbaren vertikalen Bereich
				
				-- Zeile beginnt
				if EdgeDet.risingEdge(hrefBuf) then
					isLine := True; -- sichtbarer Bereich horizontal setzen
				--  
				elsif EdgeDet.fallingEdge(hrefBuf) then
					isLine := False; -- sichtbarer Bereich horizontal ruecksetzen
				end if;
				
				-- Datenspeicherung, wenn sichtbar horizontal und Kameratakt
				if isLine and EdgeDet.risingEdge(pclkBuf) then
				-- Position des Bildpunktes im 32 Bit SRAM-Wert
					case packPos is
						when 0 =>
							writeDataTmp(7 downto 0) := dinBufPrev;
						when 1 =>
							writeDataTmp(15 downto 8) := dinBufPrev;
						when 2 =>
							writeDataTmp(23 downto 16) := dinBufPrev;
						when 3 =>
							writeDataTmp(31 downto 24) := dinBufPrev;
							
							writeData <= writeDataTmp; -- Daten weiterreichen
			
							writeRequest <= True; -- 32 Bit Wert komplett -> Abspeichern 
							
						when others => null;
					end case;
					packPos <= packPos + 1; -- Position fuer Bildpunkt weiterschieben
								
				end if;
			end if;
			
			-- Abtasten der Signale von cam1 oder cam2
			dinBufPrev <= dinBuf;
			case sw(7) is -- Auswahl der Kamera  
				when '0' =>
					dinBuf <= D1(9 downto 2); -- Uebernahme des Bildpunktes (nur 8 Bit)
					EdgeDet.setNext(pclkBuf, PCLK1);
					EdgeDet.setNext(hrefBuf, HREF1);
					EdgeDet.setNext(vsyncBuf, camVSYNC1);
				when others =>
					dinBuf <= D2(9 downto 2); -- Uebernahme des Bildpunktes (nur 8 Bit)
					EdgeDet.setNext(pclkBuf, PCLK2);
					EdgeDet.setNext(hrefBuf, HREF2);
					EdgeDet.setNext(vsyncBuf, camVSYNC2);
				end case;
			
			if pictCnt = pictRef then
				sigledBuf <= not sigledBuf;
			end if;
			
		end if;
	end process;



	-- ###########################################
	-- ### SCCB bus / camera control
	-- ###########################################
	
	rs232TX <= rsrcv_int when rs232_used = '1' else rs232TX_int;
	rs232RX_int <= rsrcv_int when rs232_used = '1' else rs232RX;

	CamInit : entity CamInit(Behavioral) port map (
		clk => clk1x,
		cam_init => cam_init,
		rs232_used => rs232_used,
		rsrcv_int =>rsrcv_int
	);
	
	CamControl_Inst : entity CamControl(arch) port map (
		clk => clk1x,
		rs232RX => rs232RX_int,--rs232RX,
		rs232TX => rs232TX_int,
		RSTB => RSTB,
		PWDN => PWDN,
		SCL => SCL,
		SDA => SDABuf,
		reset => captureReset
	);
	
	XCLK <= clkxclk(3); -- Takt fuer Kamera-Chip erzeugen (ca 6 MHz)
	
	-- sende RS232-daten an beide Kameras
	SDA1 <= SDABuf;
	SDA2 <= SDABuf;

	
end Behavioral;

